@@ -164,13 +164,13 @@ class KobertBiGRUCRF(nn.Module):
 
 class KobertSequenceFeatureExtractor(nn.Module):
     """ koBERT alone """
-    def __init__(self, config, num_classes) -> None:
+    def __init__(self, config=None, num_classes=0) -> None:
         super(KobertSequenceFeatureExtractor, self).__init__()
         self.bert, self.vocab = get_pytorch_kobert_model()
         self.dropout = nn.Dropout(config.dropout)
         self.position_wise_ff = nn.Linear(config.hidden_size, num_classes)
 
-    def forward(self, input_ids, token_type_ids=None):
+    def forward(self, input_ids, token_type_ids=None, tags=None):
         attention_mask = input_ids.ne(self.vocab.token_to_idx[self.vocab.padding_token]).float()
         outputs = self.bert(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask)
         last_encoder_layer = outputs[0]
